#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "UDCF.h"
#include "UDCFSupport.h"

/* ---- Unix environment defines and includes ---- */
#ifndef WIN32

#include "config.h"

#ifdef WITH_DMALLOC 
#include<dmalloc.h>
#endif /* WITH_DMALLOC */

#endif /* WIN32 */

/* ---- Anything absolutely smaller then 'epsilon' is equal 0.0 ---- */
const double epsilon = 1.0e-5;  

/* ---- File handle for scanner ---- */
extern FILE *yyin;

/* ---- Some prototypes for scanner/parser functions ---- */
extern void yyrestart(FILE*);
extern void resetLineNo();

#ifdef HAVE_FLEX
extern void yyfreebuffer();
#endif

#ifndef YYPARSE_RETURN_TYPE
#define YYPARSE_RETURN_TYPE int
#endif
extern YYPARSE_RETURN_TYPE yyparse ();

/* ---- Global pointer for complete data contained in a UDCF file ---- */
extern UDCF* UDCFGlobal; 

/* ---- Number of columns in a line of section <SAMPLE>  </SAMPLE> ---- */
#define COLUMNS 6 

void __API UDCFInit(UDCF* data)
{
 /* ---- Init the dynamic array of groups ---- */
 data -> groupSize  = 0;
 data -> groupIndex = -1;
 data -> groupList  = (UDCFGroup *) NULL;

 /* ---- Init all strings ---- */
 data -> vendor       = (char *) NULL;
 data -> model        = (char *) NULL;
 data -> personalInfo = (char *) NULL;
 data -> serialID     = (char *) NULL;

 /* ---- Init Version and number of dives ---- */
 data -> driverVersion = (long) C_DIVE_INVALID;
 data -> totalDives = (long) C_DIVE_INVALID;

}

UDCF* __API UDCFRead(FILE *file)
{

 /* ---- Read and parse the file and return a UDCF-struct or an invalid pointer ---- */
 return UDCFParse(file);

}

UDCF* __API UDCFReadFile(char *fileName)
{
 UDCF* profile = (UDCF*) NULL;
 FILE* file;
 
 /* ---- Open the file ---- */
#ifdef WIN32
 file = fopen(fileName,"rt");
#else
 file = fopen(fileName,"r");
#endif 

 /* ---- Check the fopen-call ---- */
 if (file != (FILE *) NULL) {
  
  /* ---- Parse the file ---- */
  profile = UDCFRead(file);
  
  /* ---- Close the file ---- */             
  fclose(file);
 
 }
 
 /* ---- Return a UDCF-struct or an invalid pointer ---- */
 return profile;
}

UDCF* __API UDCFParse(FILE* file)
{
 UDCF* data = (UDCF*) NULL;

 /* ---- Allocate memory and test success ---- */
 data = malloc(sizeof(UDCF));

 /* ---- Check the creation ---- */
 if (data == (UDCF*)NULL) {

  /* ---- Tell an error and quit ---- */
  fprintf(stderr,"Can't allocate memory for UDCF data\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Init the UDCF-struct ---- */
 UDCFInit(data);

 /* ---- Set the lexer datastream ---- */
 yyin = file;

 /* ---- Reset the scanner and file buffers ---- */
 resetLineNo();
 yyrestart(file);

 /* ---- Start parsing and set global pointer correctly ---- */
 UDCFGlobal = data;
 if (yyparse() == 0) 
  
  /* ---- Return a valid data-block ---- */
  return data;

 else {
 
  /* ---- Free memory block an return 'NULL' ---- */
  UDCFFree(data);
  return NULL;
 
 }

}

void __API UDCFFreeMixes(UDCFDive* dive)
{
 long i;

 /* ---- Check for valid data ---- */
 if (dive == (UDCFDive *) NULL) return;

 /* ---- Free all names of mixes ---- */
 for (i = 0; i <= dive -> mixIndex; i++) {

  /* ---- Free the name ---- */
  free(dive -> mixList[i].mixName);

 } 

 /* ---- Now free the array and init the variables ---- */
 free(dive -> mixList);
 dive -> mixList = (UDCFMix *) NULL;
 dive -> mixSize = 0;
 dive -> mixIndex = -1;

}

void __API UDCFFreeDive(UDCFDive* dive)
{
#ifndef NDEBUG
 fprintf(stderr, "Free a dive\n");
#endif

 /* ---- Free the mixes list ---- */
 UDCFFreeMixes(dive);

 /* ---- Free the Meta-Data field if needed ---- */
 if (dive -> prgData != (char *) NULL) free(dive -> prgData);

 /* ---- Free the place-data field if needed ---- */
 if (dive -> place != (char *) NULL) free(dive -> place);

 /* ---- Free all samples ---- */
 free(dive -> sampleList);

}

void __API UDCFFreeDiveAll(UDCFDive* dive)
{
#ifndef NDEBUG
 fprintf(stderr, "Free a dive completely\n");
#endif

 /* ---- Free all data hold by the dive ---- */
 UDCFFreeDive(dive);

 /* ---- Free the dive-struct itself ---- */
 free(dive);

}

void __API UDCFFreeGroup(UDCFGroup* group)
{
 int i;

#ifndef NDEBUG
 fprintf(stderr, "Free a group\n");
#endif
 
 /* ---- Free all dives ---- */
 for (i = 0; i <= group -> diveIndex; i++) {

  /* ---- Free the ith dive ---- */ 
  UDCFFreeDive(&(group -> diveList[i]));

 }
 
 /* ---- Free the dynamic array ---- */
 free(group -> diveList);

}

void __API UDCFFree(UDCF* data)
{
 int i;

#ifndef NDEBUG
 fprintf(stderr, "Free the dive\n");
#endif

 /* ---- Free all groups of dives ---- */
 for (i = 0; i <= data -> groupIndex; i++) {

  /* ---- Free the ith group ---- */ 
  UDCFFreeGroup(&(data -> groupList[i]));

 }

 /* ---- Free array of groups ---- */
 if (data -> groupList != (UDCFGroup *) NULL) free(data -> groupList);
 
 /* ---- Free all copy's of strings ---- */
 if (data -> vendor       != (char *) NULL) free(data -> vendor);
 if (data -> model        != (char *) NULL) free(data -> model);
 if (data -> personalInfo != (char *) NULL) free(data -> personalInfo);
 if (data -> serialID     != (char *) NULL) free(data -> serialID);

 /* ---- Free main data block ---- */
 free(data);

#ifdef HAVE_FLEX
 /* ---- Free memory used by flex ---- */
 yyfreebuffer();
#endif

}

/* ---- Some prototypes ---- */
bool UDCFWriteGroup(FILE* fileName, UDCFGroup* group);
bool UDCFWriteDive(FILE* fileName, UDCFDive* dive);

bool __API UDCFWrite(FILE* file, UDCF* data)
{
 long gInd;

 /* ---- Check if valid ---- */
 if (data == (UDCF*) NULL) return false;

 /* ---- Write version information ---- */
 fprintf(file, "<PROFILE UDCF=\"%ld\">\n", data -> version);

 /* ---- Write units used ---- */
 if (data -> measure == metric) {
 
  /* ---- Metric units ---- */
  fprintf(file, "<UNITS>Metric</UNITS>\n");
 
 } else if (data -> measure == imperial) {

  /* ---- Imperial units ---- */
  fprintf(file, "<UNITS>Imperial</UNITS>");

 } else {

  /* ---- This should never happen ---- */
  fprintf(file, "Invalid units marker\n");
  return false;

 }

 /* ---- Write device section ---- */
 fprintf(file,"<DEVICE>\n");
 fprintf(file," <VENDOR>%s</VENDOR>\n", data -> vendor);
 fprintf(file," <MODEL>%s</MODEL>\n", data -> model);
 fprintf(file," <VERSION>%ld</VERSION>\n", data -> driverVersion);
 fprintf(file,"</DEVICE>\n");

 /* ---- Check for valid additional info section ---- */
 if ((data -> personalInfo != (char*) NULL) ||
     (data -> totalDives != C_DIVE_INVALID) ||
     (data -> serialID != (char*) NULL)) {

  /* ---- Write enties of additional info section (if needed) ---- */
  fprintf(file,"<ADDINFO>\n"); 
  if (data -> personalInfo != (char*) NULL) 
   fprintf(file," <PERSINFO>%s</PERSINFO>\n", data ->  personalInfo);
  if (data -> totalDives != C_DIVE_INVALID)
   fprintf(file," <TOTALDIVES>%ld</TOTALDIVES>\n", data -> totalDives);
  if (data -> serialID != (char*) NULL)
   fprintf(file," <SERIALID>%s</SERIALID>\n", data -> serialID);
  fprintf(file,"</ADDINFO>\n");

 }

 /* ---- Write all groups of dives ---- */
 for (gInd = 0; gInd <= data -> groupIndex; gInd++) { 

  /* ---- Write the 'gInd'th group ---- */
  if (!UDCFWriteGroup(file, &(data -> groupList[gInd]))) return false;

 }

 /* ---- Write end - of marker ---- */
 fprintf(file, "</PROFILE>\n");

 /* ---- No errors occurred ---- */
 return true;

}

bool __API UDCFWriteFile(char* fileName, UDCF* data)
{
 FILE* file;
 bool result = false;

 /* ---- Open the file for writing ---- */
#ifdef WIN32
 file = fopen(fileName,"wt");
#else
 file = fopen(fileName,"w");
#endif 

 /* ---- Check the fopen-call ---- */
 if (file != (FILE *) NULL) {
  
  /* ---- Write the file ---- */
  result = UDCFWrite(file, data);
  
  /* ---- Close the file ---- */             
  fclose(file);
 
 }

 /* ---- Return the result ---- */
 return result;

}

bool UDCFWriteGroup(FILE* file, UDCFGroup* group)
{
 long dInd;

 /* ---- Write Marker ---- */
 fprintf(file, "<REPGROUP>\n");

 /* ---- Write all dives ---- */ 
 for (dInd = 0; dInd <= group -> diveIndex; dInd++) {

  /* ---- Write the 'dInd'th dive ---- */
  if (!UDCFWriteDive(file, &(group -> diveList[dInd]))) return false;

 }

 /* ---- Write End Marker ---- */
 fprintf(file, "</REPGROUP>\n");

 /* ---- Return with no error ---- */
 return true;

}

bool UDCFWriteDive(FILE* file, UDCFDive* dive)
{
 long sInd,j,i;
 bool wasSpecial;

 /* ---- Write dive marker ---- */
 fprintf(file, " <DIVE>\n");

 /* ---- Check for valid place ---- */
 if (dive -> place != (char *) NULL) {

  /* ---- Write the data ---- */
  fprintf(file, "  <PLACE>");
  fprintf(file, "%s", dive -> place);
  fprintf(file, "  </PLACE>\n");

 }

 /* ---- Write date/time information about 'dive' ---- */
 fprintf(file, "  <DATE><YEAR>%2.2d</YEAR><MONTH>%2.2d</MONTH><DAY>%2.2d</DAY></DATE>\n", 
               dive -> year, 
               dive -> month, 
               dive -> day);  
 fprintf(file, "  <TIME><HOUR>%2.2d</HOUR><MINUTE>%2.2d</MINUTE></TIME>\n", 
               dive -> hour, 
               dive -> minute);

 /* ---- Check for a valid surface interval ---- */
 if (dive -> surfaceInterval == C_NAI)  
  fprintf(file, "  <SURFACEINTERVAL>infinity</SURFACEINTERVAL>\n"); 
 else
  fprintf(file, "  <SURFACEINTERVAL>%6.2f</SURFACEINTERVAL>\n", 
                dive -> surfaceInterval); 

 /* ---- Write rest of dive information ---- */
 if (dive -> temperature != C_DIVE_INVALID)
  fprintf(file, "  <TEMPERATURE>%6.2f</TEMPERATURE>\n", dive -> temperature); 
 fprintf(file, "  <DENSITY>%6.2f</DENSITY>\n", dive -> density);
 fprintf(file, "  <ALTITUDE>%6.2f</ALTITUDE>\n", dive -> altitude);

 /* ---- Write the GASES section marker ---- */
 fprintf(file, "  <GASES>\n");

 /* ---- Check for valid mixes ---- */
 if (dive -> mixList != (UDCFMix *) NULL) {

  /* ---- Iterate over all mixes ---- */
  for (i = 0; i <= dive -> mixIndex; i++) {

   /* ---- Write the i'th mix ---- */
   fprintf(file, "   <MIX>\n");
   fprintf(file, "    <MIXNAME>%s</MIXNAME>\n", dive -> mixList[i].mixName);

   /* ---- Check for existing tank-info ---- */
   if (dive -> mixList[i].pressureStart != C_DIVE_INVALID) {

    /* ---- Write start-tag of tank-info ---- */
    fprintf(file, "    <TANK>\n");

    /* ---- Check for valid tank-volume-info ---- */
    if (dive -> mixList[i].tankVolume != C_DIVE_INVALID)
     fprintf(file, "     <TANKVOLUME>%5.3f</TANKVOLUME>\n", dive -> mixList[i].tankVolume);

    /* ---- Write tank-pressure at the beginning ---- */
    fprintf(file, "     <PSTART>%6.2f</PSTART>\n", dive -> mixList[i].pressureStart);
  
    /* ---- Check for end-pressure ---- */
    if (dive -> mixList[i].pressureEnd != C_DIVE_INVALID)
     fprintf(file, "     <PEND>%6.2f</PEND>\n", dive -> mixList[i].pressureEnd);
 
    /* ---- Terminate tank-info section ---- */
    fprintf(file, "    </TANK>\n");
 
   }

   fprintf(file, "    <O2>%5.3f</O2>\n", dive -> mixList[i].o2Ratio);
   fprintf(file, "    <N2>%5.3f</N2>\n", dive -> mixList[i].n2Ratio);
   fprintf(file, "    <HE>%5.3f</HE>\n", dive -> mixList[i].heRatio);
   fprintf(file, "   </MIX>\n");

  }  

 } else {

  /* ---- Write an invalid mixes section (Should never happen) ---- */
  fprintf(file, "   <MIX>\n");
  fprintf(file, "    <MIXNAME>UnknownMix</MIXNAME>\n");
  fprintf(file, "    <O2>%5.3f</O2>\n", 0.0);
  fprintf(file, "    <N2>%5.3f</N2>\n", 1.0);
  fprintf(file, "    <HE>%5.3f</HE>\n", 0.0);
  fprintf(file, "   </MIX>\n");

 }

 /* ---- Write the end-of-GASES-section marker ---- */
 fprintf(file, "  </GASES>\n");


 /* ---- Do we have application specific meta data ? ---- */
 if (dive -> prgData != (char *) NULL) {

  /* ---- Write the data ---- */
  fprintf(file, "  <PROGRAM>");
  fprintf(file, "%s", dive -> prgData);
  fprintf(file, "  </PROGRAM>\n");

 }

 /* ---- Write the mode ---- */
 if (dive -> mode == M_DELTA) {

  /* ---- Insert the delta mode marker ---- */
  fprintf(file, "  <DELTAMODE/>\n"); 
  
 } else {

  /* ---- Simply insert the tag for time/depth-mode ---- */
  fprintf(file, "  <TIMEDEPTHMODE/>\n");


 }

 /* ---- Write samples header ---- */
 fprintf(file, "  <SAMPLES>");
 if (dive -> mode == M_DELTA) fprintf(file, "\n   "); 

 /* ---- Init column counter and flag signalling ---- */
 /* ---- that a special marker was written       ---- */ 
 j = 0;
 wasSpecial = false;

 /* ---- Write all samples ---- */
 for (sInd = 0; sInd <= dive -> sampleIndex; sInd++) {

  /* ---- Check if last symbol was a special marker ---- */
  if (wasSpecial) fprintf(file, "   ");

  /* ---- Check if we need a new line ---- */
  if (((j % COLUMNS) == 0) && (j > 0)) fprintf(file, "\n   ");

  /* ---- Check for special samples ----- */
  if ((dive -> sampleList[sInd].type == ST_MARK) &&
      (dive -> sampleList[sInd].data.mark == C_SURF)) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "Surface" marker ---- */
   fprintf(file, "<ALARM>SURFACE</ALARM>\n");

   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if (dive -> sampleList[sInd].type == ST_TEMPERATURE) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "TEMPERATURE" marker ---- */
   fprintf(file, 
           "<TEMPERATURE>%6.2f</TEMPERATURE>\n", 
           dive -> sampleList[sInd].data.temperature);

   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if (dive -> sampleList[sInd].type == ST_MIX) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "SWITCH" marker ---- */
   fprintf(file, "<SWITCH>%s</SWITCH>\n", 
                 dive -> mixList[dive -> sampleList[sInd].data.mix].mixName); 
  
   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if (dive -> sampleList[sInd].type == ST_WORK) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "SWITCH" marker ---- */
   fprintf(file, "<WORK>%6.2f</WORK>\n", dive -> sampleList[sInd].data.work); 
            
   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if ((dive -> sampleList[sInd].type == ST_SAMPLETIME) && 
             (dive -> mode == M_DELTA)) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");
  
   /* ---- Write the delta-marker ---- */
   fprintf(file, "<DELTA>%6.2f</DELTA>\n", 
                 dive -> sampleList[sInd].data.sampleTime);

   /* ---- Reset column-counter ---- */    
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if ((dive -> sampleList[sInd].type == ST_MARK) &&
             (dive -> sampleList[sInd].data.mark == C_DECO)) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "Deco" marker ---- */
   fprintf(file, "<ALARM>DECO</ALARM>\n");
  
   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if ((dive -> sampleList[sInd].type == ST_MARK) &&
             (dive -> sampleList[sInd].data.mark == C_BREATH)) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print "Deco" marker ---- */
   fprintf(file, "<ALARM>BREATH</ALARM>\n");
  
   /* ---- Reset column-counter ---- */
   j = -1;

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if ((dive -> sampleList[sInd].type == ST_MARK) &&
             (dive -> sampleList[sInd].data.mark == C_ASC)) {

   /* ---- Check if we need a linefeed ---- */
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n   ");

   /* ---- Print slow marker ---- */
   fprintf(file, "<ALARM>ASCENT</ALARM>\n");

   /* ---- Reset column-counter ---- */
   j = -1;
   
   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = true;

  } else if ((dive -> sampleList[sInd].type == ST_DEPTH)) {

   /* ---- Print depth ---- */
   fprintf(file, "<D>%6.2f</D>", dive -> sampleList[sInd].data.depth);

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = false;

  } else if ((dive -> sampleList[sInd].type == ST_TD)) {

   /* ---- Print depth ---- */
   fprintf(file, "<T>%6.2f</T><D>%6.2f</D>", 
                 dive -> sampleList[sInd].data.point.x,
                 dive -> sampleList[sInd].data.point.y);

   /* ---- Set flag that we wrote a special marker ---- */
   wasSpecial = false;

  }

  /* ---- Check if we had already printed the last sample ---- */
  if (sInd == dive -> sampleIndex) {

   /* ---- Check if we have to add a new line ---- */ 
   if (((j % COLUMNS) != 0) && (!wasSpecial)) fprintf(file, "\n");

  } 

  /* ---- Inc column-counter ---- */
  j++;

 }

 /* ---- Write end of samples header ---- */
 fprintf(file, "  </SAMPLES>\n"); 

 /* ---- Write end of dive marker ---- */
 fprintf(file, " </DIVE>\n");

 /* ---- No error occured ---- */
 return true;

}

UDCFDive* __API UDCFConvertDelta2TD(UDCFDive* dive)
{
 long i = 0;
 long j = 0;
 long ndt = 1;
 bool insertZero = false;
 UDCFDive* newDive;
 double sampleTime = 0.0;

 /* ---- Check for valid dive ---- */
 if (dive == (UDCFDive *) NULL) return (UDCFDive *) NULL;

 /* ---- Allocate memory for dive - struct ---- */
 newDive = malloc (sizeof(UDCFDive));

 /* ---- Check the allocation ---- */
 if (newDive == (UDCFDive *) NULL) {

  /* ---- Give an error and terminate ---- */
  fprintf(stderr, "Can't allocate memory for UDCFConvertDelta2TD\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Init the new dive ---- */
 InitDive(newDive);

 /* ---- Copy the entries of the dive-struct ---- */
 *newDive = *dive;
 newDive -> sampleList = (UDCFSample *) NULL;

 /* ---- Check if we have to copy the place ---- */
 if (dive -> place != (char *) NULL) {

  /* ---- Make a copy of the diving place data ---- */
  newDive -> place = strdup(dive -> place);

  /* ---- Check the copy ---- */
  if (newDive -> place == (char *) NULL) {

   /* ---- Give an error and terminate ---- */
   fprintf(stderr, "Can't allocate memory for place description\n");
   exit(EXIT_FAILURE);

  }

 }

 /* ---- Check if we have to make a copy of the prgData field ---- */
 if (dive -> prgData != (char *) NULL) {

  /* ---- Make a copy of the prgData field ---- */
  newDive -> prgData = strdup(dive -> prgData);

  /* ---- Check the copy ---- */
  if (newDive -> prgData == (char *) NULL) { 

   /* ---- Give an error and terminate ---- */
   fprintf(stderr, "Can't allocate memory for external program data\n");
   exit(EXIT_FAILURE);

  }

 }

 /* ---- Allocate memory for mix-array ----- */
 newDive -> mixList = malloc (dive -> mixSize * sizeof(UDCFMix));

 /* ---- Check the allocation ---- */
 if (newDive -> mixList == (UDCFMix *) NULL) {

  /* ---- Give an error and terminate ---- */
  fprintf(stderr, "No memory for mix-array in UDCFConvertDelta2TD\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Make copy of mix-array ---- */
 (void) memcpy((void *) newDive -> mixList, 
               (void *) dive -> mixList, 
               newDive -> mixSize * sizeof(UDCFMix)); 

 /* ---- Clone all mixNames ---- */
 for (i = 0; i <= dive -> mixIndex; i++) {

  /* ---- Clone the i'th mixname ---- */
  newDive -> mixList[i].mixName = strdup(dive -> mixList[i].mixName);

  /* ---- Check for successful cloning ---- */
  if (newDive -> mixList[i].mixName == (char *) NULL) {

   /* ---- Give an error and terminate ---- */
   fprintf(stderr, "Can not clone mixname in UDCFConvertDelta2TD\n");
   exit(EXIT_FAILURE);

  }  

 }

 /* ---- Check if already converted ---- */
 if (dive -> mode == M_TD) {
 
  /* ---- Allocate memory for 'sampleList' ---- */
  newDive -> sampleList = malloc((newDive -> sampleSize) * sizeof(UDCFSample));

  /* ---- Check the allocation ---- */
  if (newDive -> sampleList == (UDCFSample *) NULL) {

   /* ---- Give an error and terminate ---- */
   fprintf(stderr, "Can't allocate memory for samplelist in UDCFConvertDelta2TD\n");
   exit(EXIT_FAILURE);

  }

  /* ---- Copy the data ---- */
  (void) memcpy((void *) newDive -> sampleList, 
                (void *) dive -> sampleList,
                newDive -> sampleSize * sizeof(UDCFSample));


  /* ---- Return the copied dive ---- */
  return newDive;

 }

 /* ---- Convert dive to time/depth-mode => Look if first depth is equal 0.0 ---- */
 i = 0; insertZero = false;
 while (i <= dive -> sampleIndex) {

  /* ---- Check if sample is a depth and exit loop ---- */
  if (dive -> sampleList[i].type == ST_DEPTH) {

   /* ---- Check if depth is != 0.0 ---- */
   if (fabs(dive -> sampleList[i].data.depth) > epsilon) insertZero = true; 

   /* ---- We found a sample -> exit loop ---- */
   break;

  }

  /* ---- Check for sampletime (Needed: We have to initialize 'sampleTime') ---- */
  if (dive -> sampleList[i].type == ST_SAMPLETIME) 
   sampleTime = dive -> sampleList[i].data.sampleTime;

  /* ---- Check next sample ---- */
  i++;

 }

 /* ---- Set the new mode ---- */
 newDive -> mode = M_TD;

 /* ---- Check if we must insert a point at (0|0) ---- */
 if (insertZero) { 

  /* ---- Enlarge the sample array by 1 ---- */
  newDive -> sampleSize++;
  newDive -> sampleIndex++;

 }

 /* ---- Allocate the memory for the new array ---- */
 newDive -> sampleList = malloc((newDive -> sampleSize) * sizeof(UDCFSample));

 /* ---- Check the creation ---- */
 if (newDive -> sampleList == (UDCFSample *) NULL) {

  /* ---- Give an error and terminate ---- */
  fprintf(stderr, 
          "Can't allocate memory for converted samples in UDCFConvertDelta2TD\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Start at the beginning of the dynamic array again ---- */
 i = 0;
 
 /* ---- Look for first depth sample and copy all others ---- */
 while ((i <= dive -> sampleIndex) && (dive -> sampleList[i].type != ST_DEPTH)) {

  /* ---- Copy the sample ---- */
  newDive -> sampleList[i] = dive -> sampleList[i];

  /* ---- Check next sample ---- */
  i++;

 }

 /* ---- Let j be the index for the array in 'newDive' ---- */
 j = i;

 /* ---- Check if we must insert a point at (0|0) ---- */
 if (insertZero) { 

  /* ---- Insert a point at (0|0) ---- */
  newDive -> sampleList[j].data.point.x = (double) 0.0;
  newDive -> sampleList[j].data.point.y = (double) 0.0;
  newDive -> sampleList[j].type = ST_TD;

  /* ---- Go to next sample (Note: j == i + 1 from now) ---- */
  j++;

 }

 /* ---- Copy rest of samples ---- */
 ndt = 1;
 while (i <= dive -> sampleIndex) {

  /* ---- Check for depth-sample ---- */
  if (dive -> sampleList[i].type == ST_DEPTH) {

   /* ---- Calculate the new sample ---- */
   newDive -> sampleList[j].data.point.x = (ndt * sampleTime) / 60.0;
   newDive -> sampleList[j].data.point.y = dive -> sampleList[i].data.depth;
   newDive -> sampleList[j].type = ST_TD;

   /* ---- Increment time interval counter ---- */
   ndt++;

  } else if (dive -> sampleList[i].type == ST_SAMPLETIME) {

   /* ---- Set the new sampleTime and copy the marker ---- */
   sampleTime = dive -> sampleList[i].data.sampleTime;
   newDive -> sampleList[j] = dive -> sampleList[i];

  } else {

   /* ---- Simply copy the marker / gas switch / temperature ---- */
   newDive -> sampleList[j] = dive -> sampleList[i];

  }

  /* ---- Go to next sample (In both dynamic array's holding the samples) ---- */
  i++; j++;

 }
 
 /* ---- At the surface with last sample? (Let j be the index of the last sample) ---- */
 i--;
 if (!((dive -> sampleList[i].type == ST_MARK) &&
       (dive -> sampleList[i].data.mark == C_SURF)) || 
      ((dive -> sampleList[i].type == ST_DEPTH) &&
       (fabs(dive -> sampleList[i].data.depth) > epsilon))) {

  /* ---- Resize the array ---- */
  newDive -> sampleSize++;
  newDive -> sampleIndex++;
  newDive -> sampleList = realloc(newDive -> sampleList,
                                  newDive -> sampleSize * sizeof(UDCFSample));

  /* ---- Check the creation ---- */
  if (newDive -> sampleList == (UDCFSample *) NULL) {

   /* ---- Give an error and terminate ---- */
   fprintf(stderr, "Can't add surface marker in UDCFConvertDelta2TD\n");
   exit(EXIT_FAILURE);

  }

  /* ---- Add a surface marker ---- */
  newDive -> sampleList[j].type = ST_MARK;
  newDive -> sampleList[j].data.mark = C_SURF;  

 } 

#ifndef NDEBUG
 fprintf(stderr, "-----------\n");
 PrintDive(dive);
 fprintf(stderr, "-----------\n");
 PrintDive(newDive);
 fprintf(stderr, "-----------\n");
#endif

 /* ---- Return the converted dive ---- */
 return newDive;

}

void __API UDCFGetDiveData(UDCFDive* dive, double* time, double* depth)
{
 bool firstChanged = true;
 long i;
 double sampleTime = 0.0;
#ifndef NDEBUG
 double maxTime = (double) 0.0;
 bool timeDepthMode = true;
#endif

 /* ---- Init depth ---- */
 *depth = (double) 0.0;

 /* ---- Check all entries ---- */
 for (i = 0; i <= dive -> sampleIndex; i++) {

  /* ---- Check for new sample-time ---- */
  if (dive -> sampleList[i].type == ST_SAMPLETIME) {

   /* ---- Set the new sample time ---- */
   sampleTime = dive -> sampleList[i].data.sampleTime;

   /* ---- First time to change the sample-time? ---- */
   if (firstChanged) {
  
    /* ---- Set the time (Start at time = 0.0 sec when valid sample is found) ---- */
    *time = -1.0 * sampleTime;
  
    /* ---- Never reset 'time' ---- */
    firstChanged = false;

   }

  }

  /* ---- Check for a depth-sample ---- */
  if (dive -> sampleList[i].type == ST_DEPTH) {

#ifndef NDEBUG
   /* ---- Remember sample-mode ---- */
   timeDepthMode = false;
#endif

   /* ---- Check whether depth has increased ---- */
   if (dive -> sampleList[i].data.depth > *depth) 
    *depth = dive -> sampleList[i].data.depth;

   /* ---- Increase time of dive ---- */
   *time += sampleTime;

  }

  /* ---- Check for x/y-sample ---- */
  if (dive -> sampleList[i].type == ST_TD) {

#ifndef NDEBUG
   /* ---- Remember sample-mode ---- */
   timeDepthMode = true;
#endif
 
   /* ---- Check whether depth has increased ---- */
   if (dive -> sampleList[i].data.point.y > *depth) 
    *depth = dive -> sampleList[i].data.point.y;

#ifndef NDEBUG
   maxTime = dive -> sampleList[i].data.point.x;
#endif 

   /* ---- Increase time of dive ---- */
   *time += sampleTime;
   
  }

 }

 /* ---- Set time to minutes ---- */
 *time /= 60.0;

#ifndef NDEBUG
 if ((fabs(maxTime - *time) > epsilon) && timeDepthMode) {
  fprintf(stderr, "Wrong dive-time calculated! Found: "); 
  fprintf(stderr, "%f calculated %f\n", maxTime, *time);
 }
#endif

}

/* ---- Calculate maximal depth and and divetime of 'dive' ---- */
double __API UDCFAverageDepth(UDCFDive* dive)
{
 double avgDepth = (double) 0.0;
 long samples = 0;
 long i;

 /* ---- Check all entries ---- */
 for (i = 0; i <= dive -> sampleIndex; i++) {

  /* ---- Check for a depth-sample ---- */
  if (dive -> sampleList[i].type == ST_DEPTH) 
   avgDepth += dive -> sampleList[i].data.depth;

  /* ---- Check for a x/y-mode sample ---- */
  if (dive -> sampleList[i].type == ST_TD) 
   avgDepth += dive -> sampleList[i].data.point.y;

  /* ---- Count all samples ---- */
  samples++;

 }

 /* ---- Return the average depth or 0.0 if we have no valid average depth ---- */
 if (samples != 0) 
  return avgDepth / samples; 
 else
  return (double) 0.0;

}
