#ifndef _UCDF_
#define _UCDF_

#include <stdio.h>

#ifdef WIN32
#define __API _stdcall
#include "bool.h"
#else
/* ---- WIN32 undefined ---- */
#define __API
#ifndef __cplusplus
#include <bool.h>
#endif /* __cplusplus */
#endif /* WIN 32 */                                                              

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* ---- Anything absolutely smaller then 'epsilon' is equal 0.0 ---- */
extern const double epsilon;  

/* ---- Flag for metric/imperial units ---- */
enum measure_enum{metric,imperial};
typedef enum measure_enum UCDFMeasure;

/* ---- A structure holding a mix of He, O2 and N2 ---- */
struct udcf_mix_struct {
                        char* mixName;
                        double pressureStart, pressureEnd;
                        double tankVolume;
                        double heRatio;
                        double o2Ratio;
                        double n2Ratio;
                        double acv;
                       };
typedef struct udcf_mix_struct UDCFMix;

/* ---- Datatype for sample-mode ---- */
enum sample_mode_enum{M_DELTA, M_TD};
typedef enum sample_mode_enum UDCFModeType;

/* ---- Datatype for a point in XY-Mode ---- */
struct xy_point_struct {
                        double x;
                        double y;
                       };
typedef struct xy_point_struct UDCFPoint;

/* ---- Datatype for work-load ---- */
typedef double UDCFWork;

/* ---- Datatyps for holding a sample ---- */
enum sample_type_enum{ST_UNKNOWN = 0, 
                      ST_DEPTH, 
                      ST_TD, 
                      ST_MARK, 
                      ST_MIX, 
                      ST_SAMPLETIME, 
                      ST_TEMPERATURE, 
                      ST_WORK};
typedef enum sample_type_enum UDCFSampleType;

/* ---- Special markers for Slow, Deco, Out-of-Breath  and Surface-Mode ---- */
enum marker_type_enum{C_UNKNOWN, C_ASC, C_DECO, C_SURF, C_BREATH, C_RBT, C_ERROR};
typedef enum marker_type_enum UDCFMarkerType;

union udcf_data_union {
                       double depth;
                       double temperature;
                       double sampleTime;
                       UDCFMarkerType mark;
                       UDCFPoint point;
                       UDCFWork work;
                       long mix;
                      };
typedef union udcf_data_union UDCFSampleData;

struct udcf_sample_struct {
                           UDCFSampleType type;
                           UDCFSampleData data;
                          };
typedef struct udcf_sample_struct UDCFSample;

/* ---- Constant for an illegal time interval ---- */
#define C_NAI -1

/* ---- Constant marking invalid user data in a dive struct ---- */
#define C_DIVE_INVALID -1 

/* ---- A structure containing a single dive ---- */
struct udcf_dive_struct {
                         int day, month, year;
                         int  minute, hour;
                         double surfaceInterval;
                         char* place;
                         double density;
                         double temperature;
                         double altitude;
                         UDCFMix *mixList;
                         long mixSize;
                         long mixIndex;
                         char *prgData;
                         UDCFModeType mode;
                         UDCFSample* sampleList;
                         long sampleSize;   /* Internal: Number of available slots */
                         long sampleIndex;
                         long userDataType; /* Application specific data */
                         void *userData;    /*     "          "       "  */ 
                        };
typedef struct udcf_dive_struct UDCFDive;

/* ---- Data structure holding a group of dives ---- */
struct udcf_group_struct {
                          UDCFDive* diveList;
                          long diveSize; 
                          long diveIndex;
                         };
typedef struct udcf_group_struct UDCFGroup;

/* ---- Profile structure holding all information of a UCDF file ---- */
struct udcf_struct {
                    long version;
                    UCDFMeasure measure;
                    char* vendor;
                    char* model;
                    long driverVersion;
                    char* personalInfo;
                    long totalDives;
                    char* serialID;
                    UDCFGroup *groupList;
                    long groupSize; /* Internal: Number of available slots */
                    long groupIndex;
                   };
typedef struct udcf_struct UDCF;


/* ---- Gives the line number of a parse error ---- */
extern long errorLine;

/* ---- Init a UDCF struct ---- */
void __API UDCFInit(UDCF* data);

/* ---- Free the UDCF data ---- */
void __API UDCFFree(UDCF* data);

/* ---- Read and parse a UDCF-File from 'file' ---- */
UDCF* __API UDCFRead(FILE* file);

/* ---- Read and parse a UDCF-File from file named 'filename' ---- */
UDCF* __API UDCFReadFile(char* fileName); 

/* ---- Parse a UDCF-File 'file' and return a valid UDCF data-block ---- */
UDCF* __API UDCFParse(FILE* file);

/* ---- Write a vaild UDCF-Struct to 'file' (Return 'true' for success) ---- */
bool __API UDCFWrite(FILE* file, UDCF* data); 

/* ---- Write a UDCF-File to file named 'fileName' (Return 'true' for success) ---- */
bool __API UDCFWriteFile(char* fileName, UDCF* data); 

/* ---- Convert a dive from delta-mode into time/depth-mode ---- */
UDCFDive* __API UDCFConvertDelta2TD(UDCFDive* dive); 

/* ---- Calculate maximal depth and and divetime of 'dive' ---- */
void __API UDCFGetDiveData(UDCFDive* dive, double* time, double* depth);

/* ---- Calculate average depth of 'dive' ---- */
double __API UDCFAverageDepth(UDCFDive* dive);

/* ---- Free all data hold by the dive-struct 'dive' but not 'dive' itself ---- */
void __API UDCFFreeDive(UDCFDive* dive);

/* ---- Free a dive struct and all its data. Use it for dives created by  ---- */
/* ---- 'UDCFConvertDelta2TD'                                             ---- */
void __API UDCFFreeDiveAll(UDCFDive* dive);

#ifdef __cplusplus
}
#endif /* __cplusplus */                                                       

#endif /* _UDCF_ */
